import dash
from dash import dcc, html
import plotly.express as px
import pandas as pd
from dash.dependencies import Input, Output
import plotly.graph_objs as go
import folium

def process_map_data(factory_file, warehouse_file, od_pairs_file):
    #Read factory data
    factory_df = pd.read_csv(factory_file)
    factory_df['facility_type'] = 'Factory'

    #Read warehouse data
    warehouse_df = pd.read_csv(warehouse_file)
    warehouse_df['facility_type'] = 'Warehouse'

    #Merge factory and warehouse data
    facilities_df = pd.concat([factory_df, warehouse_df], ignore_index=True)

    #Read origin-destination pairs
    od_pairs_df = pd.read_csv(od_pairs_file)

    return facilities_df, od_pairs_df

def process_od_pairs(od_pairs_df, facilities_df):
    #Create a dictionary for mapping ID to coordinates
    id_to_coords = facilities_df.set_index('id')[['latitude', 'longitude']].to_dict('index')

    #Function to get coordinates for a given ID
    def get_coords(id):
        return id_to_coords.get(id, {'latitude': None, 'longitude': None})

    #Create columns for origin and destination coordinates
    od_pairs_df['origin_coords'] = od_pairs_df['from'].apply(lambda x: get_coords(x))
    od_pairs_df['dest_coords'] = od_pairs_df['to'].apply(lambda x: get_coords(x))

    return od_pairs_df

def process_data(factory_file, warehouse_file, od_pairs_file):
    facilities_df, od_pairs_df = process_map_data(factory_file, warehouse_file, od_pairs_file)
    od_pairs_df = process_od_pairs(od_pairs_df, facilities_df)
    return facilities_df, od_pairs_df

def read_factory_output(file_path):
    headers = ['facility_name', 'productID', 'production', 'timeStep']
    df = pd.read_csv(file_path, header=None, names=headers)
    df['production'] = df['production'].clip(lower=0)
    return df

def read_warehouse_output(file_path):
    headers = ['facility_name', 'productID', 'demand', 'inventory', 'timeStep']
    df = pd.read_csv(file_path, header=None, names=headers)
    return df

def process_chart_data(factory_file, warehouse_file):
    #Read and process the factory output file
    factory_df = read_factory_output(factory_file)

    #Read and process the warehouse output file
    warehouse_df = read_warehouse_output(warehouse_file)

    #Extract unique product IDs
    product_ids = sorted(set(factory_df['productID']).union(set(warehouse_df['productID'])))

    return factory_df, warehouse_df, product_ids

def aggregate_totals(factory_df, warehouse_df):
    #Aggregate production totals per product
    production_totals = factory_df.groupby(['facility_name', 'productID'])['production'].sum().reset_index()

    #Aggregate demand and inventory totals per product
    warehouse_totals = warehouse_df.groupby(['facility_name', 'productID'])[['demand', 'inventory']].sum().reset_index()

    return production_totals, warehouse_totals

factory_output = 'factory_output.csv'
warehouse_output = 'warehouse_output.csv'
factory_source = 'factory_source.csv'
warehouse_source = 'warehouse_source.csv'
topology_source = 'distribution_topology.csv'

# Process the data
factory_df, warehouse_df, product_ids = process_chart_data(factory_output, warehouse_output)
facilities_df, od_pairs_df = process_data(factory_source, warehouse_source, topology_source)
production_totals, warehouse_totals = aggregate_totals(factory_df, warehouse_df)

app = dash.Dash(__name__)

app.layout = html.Div([
    #Map at the top
    html.Div([
        dcc.Dropdown(
            id='map-dropdown',
            options=[{'label': pid, 'value': pid} for pid in product_ids],
            value=product_ids[0] if product_ids else None
        ),
        html.Div(id='map-container')  # Container for the Folium map
    ], style={'width': '100%', 'display': 'block'}),

    #Dropdown for the charts
    html.Div([
        dcc.Dropdown(
            id='chart-dropdown',
            options=[{'label': pid, 'value': pid} for pid in product_ids],
            value=product_ids[0] if product_ids else None
        )
    ], style={'width': '100%', 'display': 'block'}),

    #Charts
    html.Div([
        html.Div([dcc.Graph(id='chart1')], style={'width': '33%', 'display': 'inline-block'}), 
        html.Div([dcc.Graph(id='chart2')], style={'width': '33%', 'display': 'inline-block'}), 
        html.Div([dcc.Graph(id='chart3')], style={'width': '33%', 'display': 'inline-block'}) 
    ], style={'width': '100%', 'display': 'flex'})
])
def create_map_figure(selected_product_id, facilities_df, od_pairs_df, production_totals, warehouse_totals):
    # Create a map centered around an average location
    map_center = [facilities_df['latitude'].mean(), facilities_df['longitude'].mean()]
    folium_map = folium.Map(location=map_center, zoom_start=5)

    # Add facility markers
    for _, row in facilities_df.iterrows():
        facility_type = row['facility_type']
        if facility_type == 'Factory':
            # Make sure the filtering is correct
            production = production_totals[(production_totals['facility_name'] == row['name']) & (production_totals['productID'] == selected_product_id)]['production'].sum()
            popup_text = f"Name: {row['name']}\nType: Factory\nProduction: {production}"
        else:
            # Make sure the filtering is correct
            demand = warehouse_totals[(warehouse_totals['facility_name'] == row['name']) & (warehouse_totals['productID'] == selected_product_id)]['demand'].sum()
            inventory = warehouse_totals[(warehouse_totals['facility_name'] == row['name']) & (warehouse_totals['productID'] == selected_product_id)]['inventory'].sum()
            popup_text = f"Name: {row['name']}\nType: Warehouse\nDemand: {demand}, Inventory: {inventory}"

        folium.Marker(
            location=[row['latitude'], row['longitude']],
            popup=popup_text,
            icon=folium.Icon(color='blue' if facility_type == 'Factory' else 'red')
        ).add_to(folium_map)

    # Add lines for origin-destination pairs
    for _, row in od_pairs_df.iterrows():
        origin = [row['origin_coords']['latitude'], row['origin_coords']['longitude']]
        destination = [row['dest_coords']['latitude'], row['dest_coords']['longitude']]
        folium.PolyLine([origin, destination], color='green', weight=2.5, opacity=0.5).add_to(folium_map)

    return folium_map._repr_html_()  # Return the HTML representation of the map

@app.callback(
    Output('map-container', 'children'),
    [Input('map-dropdown', 'value')]
)
def update_map(selected_product_id):
    # Regenerate the map with the updated selected_product_id
    folium_map = create_map_figure(selected_product_id, facilities_df, od_pairs_df, production_totals, warehouse_totals)
    return html.Iframe(srcDoc=folium_map, width='100%', height='500px')

def create_production_chart(selected_product_id, factory_df):
    filtered_df = factory_df[factory_df['productID'] == selected_product_id]

    fig = go.Figure()

    for facility in filtered_df['facility_name'].unique():
        facility_df = filtered_df[filtered_df['facility_name'] == facility]
        fig.add_trace(go.Scatter(x=facility_df['timeStep'], y=facility_df['production'],
                                 mode='lines', name=facility))

    fig.update_layout(title='Production Over Time', xaxis_title='Time Step', yaxis_title='Production')

    return fig

def create_demand_chart(selected_product_id, warehouse_df):
    filtered_df = warehouse_df[warehouse_df['productID'] == selected_product_id]

    fig = go.Figure()

    for facility in filtered_df['facility_name'].unique():
        facility_df = filtered_df[filtered_df['facility_name'] == facility]
        fig.add_trace(go.Scatter(x=facility_df['timeStep'], y=facility_df['demand'],
                                 mode='lines', name=facility))

    fig.update_layout(title='Demand Over Time', xaxis_title='Time Step', yaxis_title='Demand')

    return fig

def create_inventory_chart(selected_product_id, warehouse_df):
    filtered_df = warehouse_df[warehouse_df['productID'] == selected_product_id]

    fig = go.Figure()

    for facility in filtered_df['facility_name'].unique():
        facility_df = filtered_df[filtered_df['facility_name'] == facility]
        fig.add_trace(go.Scatter(x=facility_df['timeStep'], y=facility_df['inventory'],
                                 mode='lines', name=facility))

    fig.update_layout(title='Inventory Over Time', xaxis_title='Time Step', yaxis_title='Inventory')

    return fig

@app.callback(
    [Output('chart1', 'figure'),
     Output('chart2', 'figure'),
     Output('chart3', 'figure')],
    [Input('chart-dropdown', 'value')]
)
def update_charts(selected_product_id):
    production_chart = create_production_chart(selected_product_id, factory_df)
    demand_chart = create_demand_chart(selected_product_id, warehouse_df)
    inventory_chart = create_inventory_chart(selected_product_id, warehouse_df)
    return production_chart, demand_chart, inventory_chart

if __name__ == '__main__':
    app.run_server(debug=True)