package supply_kit_pull_system;

import org.apache.commons.math3.random.EmpiricalDistribution;
import simudyne.core.abm.AgentBasedModel;
import simudyne.core.abm.Group;
import simudyne.core.abm.Split;
import simudyne.core.abm.supplykit.utils.LoadingBay;
import simudyne.core.annotations.ModelSettings;
import simudyne.core.data.CSVSource;
import supply_kit_pull_system.agents.Customer;
import supply_kit_pull_system.agents.MarketPlace;
import supply_kit_pull_system.agents.ProductionFacility;
import supply_kit_pull_system.agents.Warehouse;
import supply_kit_pull_system.util.Distribution;
import supply_kit_pull_system.util.ProductDetails;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

@ModelSettings(end = 100)
public class SupplyChainModel extends AgentBasedModel<Globals> {


    @Override
    public void init() {
        registerAgentTypes(ProductionFacility.class, Warehouse.class, Customer.class, MarketPlace.class);
        registerLinkTypes(Links.FactoryToWarehouseLink.class, Links.WarehouseToFactoryLink.class, Links.CustomerToMarketLink.class, Links.MarketToWarehouseLink.class, Links.WarehouseToMarketPlaceLink.class);
    }

    @Override
    public void setup() {

        getGlobals().rng = getContext().getPrng();

        CSVSource products = new CSVSource("data/products.csv");
        products.getRows().forEach(entry -> {
            String productID = entry.get("Product Code");
            double price = Double.parseDouble(entry.get("Price"));
            double volume = Double.parseDouble(entry.get("Volume"));
            double weight = Double.parseDouble(entry.get("Weight"));
            getGlobals().availableProducts.put(productID, new ProductDetails(productID, price, volume, weight));
        });

        CSVSource customerRegions = new CSVSource("data/zip1regions.csv");
        getGlobals().zipRegionData = customerRegions.getRowsFromTable();
        EmpiricalDistribution incomeDist = new Distribution().getIncomeDistribution();

        Group<MarketPlace> marketPlaceGroup = generateGroup(MarketPlace.class, 1, m -> {
            m.availableProducts = getGlobals().availableProducts;
        });

        Group<Customer> customerGroup = generateGroup(Customer.class, getGlobals().nbCustomers, customer -> {
            customer.region = getGlobals().determineCustomerRegion();
            customer.income = (int) incomeDist.sample(1)[0] * 10;
            customer.annualSalarySpend = customer.getPrng().uniform(0, getGlobals().annualSpending).sample();
            customer.availableProducts = getGlobals().availableProducts;
        });

        ConcurrentHashMap<String, Long> destAddr = new ConcurrentHashMap<>();

        CSVSource factorySource = new CSVSource("data/factory_source.csv");
        Group<ProductionFacility> factoryGroup = loadGroup(ProductionFacility.class, factorySource, fA -> {
            fA.outboundLoadingBay = new LoadingBay(5);
            fA.HEADER = "facility_name,productID,production,timeStep";
        });

        CSVSource warehouseSource = new CSVSource("data/warehouse_source.csv");
        Group<Warehouse> warehouseGroup = loadGroup(Warehouse.class, warehouseSource, fB -> {
            fB.inboundLoadingBay = new LoadingBay(5);
            destAddr.put(fB.name, fB.getID());
            fB.HEADER = "facility_name,productID,demand,inventory,timeStep";
        });

        getGlobals().warehouseLocations = destAddr;

        CSVSource distroTopology = new CSVSource("data/distribution_topology.csv");
        factoryGroup.loadConnections(warehouseGroup, Links.FactoryToWarehouseLink.class, distroTopology);

        CSVSource demandTopology = new CSVSource("data/demand_topology.csv");
        warehouseGroup.loadConnections(factoryGroup, Links.WarehouseToFactoryLink.class, demandTopology);

        warehouseGroup.fullyConnected(marketPlaceGroup, Links.WarehouseToMarketPlaceLink.class);

        customerGroup.fullyConnected(marketPlaceGroup, Links.CustomerToMarketLink.class);


        super.setup();
    }

    @Override
    public void step() {
        super.step();

        firstStep(ProductionFacility.sendInitialInventory, Warehouse.setInitialInventory);

        run(
                Split.create(
                        Customer.generateOrders,
                        Warehouse.sendInventory),
                MarketPlace.matchOrders,
                Split.create(
                        Warehouse.processOrders,
                        Customer.receiveOutOfStockMessage),
                Split.create(
                        ProductionFacility.receiveDemand,
                        Customer.receiveOrderUpdate)
        );

        run(ProductionFacility.processOutboundLoadingBay);
        run(ProductionFacility.departOutboundTransport, Warehouse.receiveInboundTransport);
        run(Warehouse.processInboundLoadingBay);

    }
}
