package supply_kit_pull_system.agents;

import simudyne.core.abm.Action;
import simudyne.core.abm.Agent;
import simudyne.core.abm.supplykit.Product;
import simudyne.core.annotations.Variable;
import supply_kit_pull_system.Globals;
import supply_kit_pull_system.Links;
import supply_kit_pull_system.util.MarketPlaceOrder;
import supply_kit_pull_system.util.Messages;
import supply_kit_pull_system.util.ProductDetails;
import supply_kit_pull_system.util.ShippingPreference;

import java.util.*;
import java.util.stream.Collectors;

public class Customer extends Agent<Globals> {
    @Variable
    public String region;
    @Variable
    public int income;
    @Variable
    public double wealth;
    @Variable
    public double annualSalarySpend;

    public Map<String, ProductDetails> availableProducts;

    public List<Messages.FullfilledOrderMessage> purchaseHistory = new ArrayList<>();

    public Map<Long, List<Product>> ordersAwaitingToArrive = new HashMap<>();


    public static Action<Customer> receiveOrderUpdate =
            Action.create(Customer.class, customer -> {
                customer.purchaseHistory = customer.getMessagesOfType(Messages.FullfilledOrderMessage.class);
            });


    //Receive any out of stock messages that were sent from the marketplace and process the refund
    public static Action<Customer> receiveOutOfStockMessage =
            Action.create(Customer.class, customer -> {
                customer.getMessagesOfType(Messages.OutOfStockMessage.class).stream().forEach(msg -> {
                    msg.outOfStockItems.entrySet().stream().forEach(item -> {
                        double refund = customer.getGlobals().availableProducts.get(item.getKey()).price;
                        customer.wealth += refund;
                    });
                });
            });


    public static Action<Customer> generateOrders =
            Action.create(Customer.class, customer -> {

                //Earn income every month
                if (customer.getContext().getTick() % 30 == 0) {
                    customer.earnIncome();
                }

                //Check if time to spend and if so generate an order
                if (customer.getPrng().uniform(0, 1).sample() <= customer.annualSalarySpend && customer.wealth > 0) {
                    customer.generateOrder();
                }
            });

    //Generate a new order for sending to the marketplace
    public void generateOrder() {

        //Calculate total spend
        double totalSpend = calculateTotalSpend();

        //Determine subset of available products to purchase and allocate total spend budget across desired products
        Map<String, Integer> allocatedSpend = allocateSpend(totalSpend, 0.0);

        //Send off the order to the marketplace (awaiting order confirmation from marketplace)
        getLinks(Links.CustomerToMarketLink.class).send(Messages.OrderMessage.class, (msg, link) -> {
            msg.marketPlaceOrderDetails = new MarketPlaceOrder(allocatedSpend, getID(), region, getContext().getTick(), getPrng().generator.nextBoolean() ? ShippingPreference.LOW_COST : ShippingPreference.EXPEDITED);
        });
    }

    //Allocate spend across a desired subset of available products
    public Map<String, Integer> allocateSpend(double totalSpend, double currentSpend) {

        // shuffle products and select a random subset of products to purchase
        List<ProductDetails> products = availableProducts.values().stream().collect(Collectors.toList());
        Collections.shuffle(products, getPrng().getRandom());
        int nbPurchases = (int) getPrng().uniform(1, getGlobals().availableProducts.size()).sample();
        List<ProductDetails> productsToPurchase = products.subList(0, nbPurchases);

        //Store prices for easier reference
        Map<String, Double> productPrices = productsToPurchase.stream().collect(Collectors.toMap(
                item -> item.productID,
                item -> item.price
        ));

        Map<String, Integer> allocation = new HashMap<>();

        //iterate over products and determine how many units of each product to purchase, monitoring current spend

        while (currentSpend <= totalSpend) {
            for (String product : productPrices.keySet()) {
                allocation.put(product, allocation.getOrDefault(product, 0) + 1);
                currentSpend += productPrices.get(product);
                if (currentSpend >= totalSpend) {
                    break;
                }
            }
        }

        //update wealth
        wealth -= currentSpend;

        return allocation;
    }

    //Calculate the total spend based on portion of income and available wealth
    public double calculateTotalSpend() {
        double spend = income / 30;
        double additionalSpend = wealth * getGlobals().annualSpending;
        return spend + additionalSpend;
    }

    //Increase wealth by monthly income
    public void earnIncome() {
        wealth += income / 12;
    }
}
