package supply_kit_pull_system.agents;


import org.apache.commons.math3.stat.regression.SimpleRegression;
import simudyne.core.abm.Action;
import simudyne.core.abm.supplykit.Facility;
import simudyne.core.abm.supplykit.Product;
import simudyne.core.abm.supplykit.utils.TransportMessages;
import simudyne.core.annotations.ModelSettings;
import simudyne.core.annotations.Variable;
import supply_kit_pull_system.Globals;
import supply_kit_pull_system.Links;
import supply_kit_pull_system.util.DemandHistory;
import supply_kit_pull_system.util.Messages;
import supply_kit_pull_system.util.Production;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

@ModelSettings(timeUnit = "MINUTES")
public class ProductionFacility extends Facility<Globals> {

    public Map<Long, List<DemandHistory>> historicalDemand = new HashMap<>();

    public Map<Long, Map<Long, List<Product>>> processedProducts = new HashMap<>();

    public Map<Long, List<Production>> productionPredictions = new HashMap<>();

    private static final String FILE_NAME = "output/" + "factory_output.csv";
    ;
    public String HEADER;


    @Variable(name = "Total Production")
    public int totalProduction;

    public static Action<ProductionFacility> processOutboundLoadingBay =
            Action.create(ProductionFacility.class, ProductionFacility::processOutboundLoadingBay);

    public static Action<ProductionFacility> departOutboundTransport =
            Action.create(ProductionFacility.class, ProductionFacility::departTransport);

    //Set initial inventory to create starting state of warehouse inventory
    public static Action<ProductionFacility> sendInitialInventory =
            Action.create(ProductionFacility.class, productionFacility -> {
                List<Product> initialShipment = new ArrayList<>();

                //Generate an initial shipment -> arrives immediately, no need to send in a truck
                productionFacility.getGlobals().availableProducts.values().stream().forEach(product -> {
                    for (int i = 0; i < productionFacility.getGlobals().initialInventory; ++i) {
                        Product newProduct = new Product(product.productID, product.weight, product.volume, "initialShipment");
                        initialShipment.add(newProduct);
                    }
                });

                //Ship of initial inventory
                productionFacility.getLinks(Links.FactoryToWarehouseLink.class).send(Messages.InitialShipmentMessage.class, (msg, link) -> {
                    msg.initialInventory = initialShipment;
                });
            });

    //Receive demand message from the warehouse and predict future production
    public static Action<ProductionFacility> receiveDemand =
            Action.create(ProductionFacility.class, productionFacility -> {
                //Store new time series for demand into the historical demand record for each product
                productionFacility.getMessagesOfType(Messages.DemandMessage.class).forEach(msg -> {
                    msg.demand.entrySet().stream().forEach(demand -> {
                        int existingStock = msg.inventory.get(demand.getKey()).size();
                        productionFacility.historicalDemand.computeIfAbsent(productionFacility.getContext().getTick(), k -> new ArrayList<>()).add(new DemandHistory(demand.getKey(), demand.getValue() - existingStock, msg.getSender()));
                    });
                });


                //Calculate future production (currently predicting demand for the next month)
                if (productionFacility.getContext().getTick() > productionFacility.getGlobals().prodPredictionTime && productionFacility.historicalDemand.size() > 0) {
                    productionFacility.calculateFutureProduction();
                    productionFacility.totalProduction = 0;
                    if (!productionFacility.productionPredictions.get(productionFacility.getContext().getTick() + 1).isEmpty()) {
                        productionFacility.totalProduction = (int) productionFacility.productionPredictions.get(productionFacility.getContext().getTick() + 1).stream().mapToDouble(Production::getProduction).sum();

                    }
                }
            });

    @Override
    protected void processOutboundLoadingBay() {
        // pull trucks from transport queue
        pullTrucksFromQueue();
        // generate products
        long processingTime = getContext().getTick() + getPrng().getNextInt(3);
        Map<Long, List<Product>> producedProductsByDestination = generateProducts();

        producedProductsByDestination.entrySet().stream().forEach(prodDest -> {
            processedProducts.computeIfAbsent(processingTime, k -> new HashMap<>()).put(prodDest.getKey(), prodDest.getValue());
        });

        if (processedProducts.entrySet().stream().filter(entry -> entry.getKey() == getContext().getTick()).findAny().isPresent()) {
            Map<Long, List<Product>> readyToDepart = new HashMap<>();
            processedProducts.entrySet().stream()
                    .filter(entry -> entry.getKey() == getContext().getTick()).forEach(readyProds -> {
                readyToDepart.putAll(readyProds.getValue());
            });
            processedProducts.get(getContext().getTick()).clear();


            if (readyToDepart != null) {
                readyToDepart.entrySet().stream().forEach(productSet -> {
                    String destination = String.valueOf(productSet.getKey());
                    Optional<TransportMessages.TransportMessage> potentialTransport = outboundLoadingBay.loadingDocks.values().stream().filter(transportMessage -> transportMessage != null && transportMessage.destination.equals(destination)).findAny();
                    if (potentialTransport.isPresent()) {
                        String trailerID = potentialTransport.get().id;
                        productSet.getValue().stream().forEach(product -> product.trailerID = trailerID);
                        potentialTransport.get().objectContents.addAll(productSet.getValue());
                    } else {
                        generateTruck(destination);
                        pullTrucksFromQueue();
                        Optional<TransportMessages.TransportMessage> t = outboundLoadingBay.loadingDocks.values().stream().filter(transportMessage -> transportMessage != null && transportMessage.destination == destination).findAny();
                        if (t.isPresent()) {
                            String trailerID = t.get().id;
                            productSet.getValue().stream().forEach(product -> product.trailerID = trailerID);
                            t.get().objectContents.addAll(productSet.getValue());
                        }
                    }
                });
            }
        }
        System.out.println("check the outbound loading bay");
    }

    @Override
    protected void departTransport() {
        List<TransportMessages.TransportMessage> departedTransport = new ArrayList<>();
        outboundLoadingBay.loadingDocks.values().stream().filter(transport -> transport != null && transport.departureTime == getContext().getTick()).forEach(t -> {
            departedTransport.add(t);
            send(TransportMessages.TransportMessage.class, tMsg -> tMsg.constructWithProducts(t.id,
                    t.transportType, t.weightCapacity, t.volumeCapacity,
                    t.weight, t.volume, t.objectContents, t.arrivalTime,
                    t.departureTime, t.origin, t.destination
            )).to(Long.parseLong(t.destination));
        });
        outboundLoadingBay.loadingDocks.entrySet().stream().forEach(bay -> {
            if (departedTransport.contains(bay.getValue())) {
                bay.setValue(null);
            }
        });
    }

    @Override
    protected void receiveInboundTransport() {
        super.receiveInboundTransport();

    }

    @Override
    protected void processInboundLoadingBay() {

    }

    public void calculateFutureProduction() {
        Map<Long, List<DemandHistory>> last30DemandByWarehouse = new HashMap<>();

        //Collect the last 30 days of historical demand
        for (Long step : historicalDemand.keySet()) {
            for (DemandHistory demandHistory : historicalDemand.get(step)) {
                long warehouseID = demandHistory.warehouseID;
                last30DemandByWarehouse.putIfAbsent(warehouseID, new LinkedList<>());
                last30DemandByWarehouse.get(warehouseID).add(demandHistory);
                if (last30DemandByWarehouse.get(warehouseID).size() > 30) {
                    last30DemandByWarehouse.get(warehouseID).remove(0);
                }
            }
        }

        last30DemandByWarehouse.entrySet().stream().forEach(entry -> {
            Map<String, List<Integer>> last30Demands = new HashMap<>();
            for (DemandHistory demandHistory : entry.getValue()) {
                String productID = demandHistory.productID;
                int demand = demandHistory.demand;
                if (demand < 0) {
                    demand = 0;
                }
                last30Demands.putIfAbsent(productID, new LinkedList<>());
                last30Demands.get(productID).add(demand);
            }

            Map<String, Double> regressionCoefficients = new HashMap<>();
            Map<String, Integer> intercepts = new HashMap<>();

            //gather regression coefficients and intercepts
            for (String productID : last30Demands.keySet()) {
                List<Integer> demands = last30Demands.get(productID);
                SimpleRegression regression = new SimpleRegression();

                for (int i = 0; i < demands.size(); i++) {
                    regression.addData(i, demands.get(i));
                }

                regressionCoefficients.put(productID, regression.getSlope());
                intercepts.put(productID, (int) regression.getIntercept());
            }

            long currentStep = getContext().getTick();

            //store future production predictions
            for (String productID : regressionCoefficients.keySet()) {
                double slope = regressionCoefficients.get(productID);
                int intercept = intercepts.get(productID);
                for (int i = 1; i <= getGlobals().prodPredictionTime; i++) {
                    long futureStep = currentStep + i;
                    int predictedDemand = calculatePredictedDemand(slope, intercept, i);
                    productionPredictions.computeIfAbsent(futureStep, k -> new ArrayList<>()).add(new Production(productID, predictedDemand, entry.getKey()));
                }
            }
        });
    }

    private int calculatePredictedDemand(double slope, int intercept, int futureStep) {
        return (int) Math.round(slope * futureStep + intercept);
    }

    // Placeholder function for how products/units enter the system
    public Map<Long, List<Product>> generateProducts() {
        Map<Long, List<Product>> newProducts = new HashMap<>();
        if (getContext().getTick() <= getGlobals().prodPredictionTime) {
            getLinks(Links.FactoryToWarehouseLink.class).stream().forEach(link -> {
                int nbNewProds = getPrng().getNextInt(5);
                for (int i = 0; i < nbNewProds; i++) {
                    getGlobals().availableProducts.values().stream().forEach(product -> {
                        Product newProduct = new Product(product.productID, product.volume, product.weight, "");
                        newProducts.computeIfAbsent(link.getTo(), k -> new ArrayList<>()).add(newProduct);
                    });
                }

            });
            System.out.println("check outbound products");
        } else {
            if (productionPredictions.containsKey(getContext().getTick())) {
                productionPredictions.get(getContext().getTick()).stream().forEach(quota -> {
                    for (int i = 0; i < quota.production; i++) {
                        String productID = quota.productID;
                        double weight = getGlobals().availableProducts.get(productID).weight;
                        double volume = getGlobals().availableProducts.get(productID).volume;
                        Product newProduct = new Product(productID, volume, weight, "");
                        newProducts.computeIfAbsent(quota.warehouseID, k -> new ArrayList<>()).add(newProduct);
                    }
                });

                for (String productID : getGlobals().availableProducts.keySet()) {
                    int production;
                    long tick = getContext().getTick();
                    production = productionPredictions.get(tick).stream().filter(prod -> prod.productID == productID).mapToInt(entry -> entry.production).sum();
                    String newData = name + "," + productID + "," + production + "," + tick;
                    writeToFile(newData);
                }
            } else {
                for (String productID : getGlobals().availableProducts.keySet()) {
                    String newData = name + "," + productID + "," + 0 + "," + getContext().getTick();
                    writeToFile(newData);
                }
            }
        }
        return newProducts;
    }

    // Placeholder function for how trucks arrive at facility
    public void generateTruck(String newDestination) {
        String trailerID = getID() + "_" + getPrng().getNextInt(1000) + "_" + newDestination;
        double weightCapacity = getPrng().uniform(10000, 25000).sample();
        double volumeCapacity = getPrng().uniform(1000, 1200).sample();

        long departureTime = getContext().getTick() + (long) getPrng().uniform(1, 10).sample();
        long arrivalTime = departureTime + (long) getPrng().uniform(1, 10).sample();

        TransportMessages.TransportMessage transportMessage = new TransportMessages.TransportMessage();
        transportMessage.constructWithProducts(trailerID,
                "truck",
                weightCapacity,
                volumeCapacity,
                0,
                0,
                new ArrayList<>(),
                arrivalTime,
                departureTime,
                name,
                newDestination);
        outboundLoadingBay.transportQueue.put(getContext().getTick(), transportMessage);
    }

    // placeholder function for handling transport queue
    public void pullTrucksFromQueue() {
        if (outboundLoadingBay.transportQueue.size() != 0) {
            List<Integer> emptyBays = findEmptyBays(outboundLoadingBay.loadingDocks);
            while (emptyBays.size() > 0 && outboundLoadingBay.transportQueue.size() > 0) {
                long nextQueuePosition = Collections.min(outboundLoadingBay.transportQueue.keySet());
                TransportMessages.TransportMessage nextTruck = outboundLoadingBay.transportQueue.get(nextQueuePosition);
                int emptyBay = emptyBays.remove(0);
                outboundLoadingBay.loadingDocks.put(emptyBay, nextTruck);
                outboundLoadingBay.transportQueue.remove(nextQueuePosition);

            }
        }
    }

    // Placeholder function for finding empty bays
    public List<Integer> findEmptyBays(Map<Integer, TransportMessages.TransportMessage> loadingDocks) {
        List<Integer> emptyBays = new ArrayList<>();
        for (Map.Entry<Integer, TransportMessages.TransportMessage> bay : loadingDocks.entrySet()) {
            if (bay.getValue() == null) {
                emptyBays.add(bay.getKey());
            }
        }
        return emptyBays;
    }

    public void writeToFile(String content) {
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(FILE_NAME, true))) {
            writer.write(content);
            writer.newLine();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
