package supply_kit_pull_system.agents;

import simudyne.core.abm.Action;
import simudyne.core.abm.supplykit.Facility;
import simudyne.core.abm.supplykit.Product;
import simudyne.core.abm.supplykit.utils.TransportMessages;
import simudyne.core.annotations.Constant;
import simudyne.core.annotations.Variable;
import supply_kit_pull_system.Globals;
import supply_kit_pull_system.Links;
import supply_kit_pull_system.util.Messages;
import supply_kit_pull_system.util.ShippingPreference;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class Warehouse extends Facility<Globals> {

    @Variable(name = "Total_Units_Processed")
    public int nbUnitsProcessed;

    @Variable(name = "Total_Shipment_Cost")
    public double totalShipmentCost;

    @Variable(name = "Cost_Per_Unit")
    public double costPerUnit;

    @Variable(name = "Total Inventory")
    public int totalInventory;

    @Constant
    public String region_name;

    @Variable(name = "Total Demand")
    public int totalDemand;

    private static final String FILE_NAME = "output/" + "warehouse_output.csv";
    ;
    public String HEADER;

    public double shipmentCostPeLb = 0.10; //placeholder value for shipping cost

    public Map<String, List<Product>> inventory = new HashMap<>();

    public static Action<Warehouse> receiveInboundTransport =
            Action.create(Warehouse.class, Warehouse::receiveInboundTransport);

    public static Action<Warehouse> processInboundLoadingBay =
            Action.create(Warehouse.class, Warehouse::processInboundLoadingBay);

    //Receive initial inventory and handle immediately
    public static Action<Warehouse> setInitialInventory =
            Action.create(Warehouse.class, warehouse -> {

                //Receive initial shipment and process into inventory
                warehouse.getMessagesOfType(Messages.InitialShipmentMessage.class).forEach(msg -> {
                    msg.initialInventory.stream().forEach(product -> {
                        warehouse.inventory.computeIfAbsent(product.productID, k -> new ArrayList<>()).add(product);
                    });
                });
            });

    //Send the current state of your inventory
    public static Action<Warehouse> sendInventory =
            Action.create(Warehouse.class, warehouse -> {
                warehouse.getLinks(Links.WarehouseToMarketPlaceLink.class).send(Messages.InventoryMessage.class, (msg, link) -> {
                    msg.inventory = warehouse.inventory;
                });
            });

    //Process incoming orders and send demand to production facility for predicting future production
    public static Action<Warehouse> processOrders =
            Action.create(Warehouse.class, warehouse -> {
                Map<String, Integer> demand = new HashMap<>();
                warehouse.totalDemand = 0;
                //loop through warehouse order msg
                warehouse.getMessagesOfType(Messages.WarehouseOrderMessage.class).forEach(order -> {
                    String productID = order.warehouseOrder.productId;
                    int nbItems = order.warehouseOrder.nbItems;
                    warehouse.totalDemand += nbItems;
                    demand.compute(productID, (k, v) -> (v == null) ? nbItems : v + nbItems);
                    List<Product> specificItems = new ArrayList<>();
                    if (warehouse.inventory.get(productID).size() < nbItems) {
                        specificItems = warehouse.inventory.get(productID);
                    } else {
                        specificItems = warehouse.inventory.get(productID).subList(0, nbItems);
                    }
                    if (specificItems.size() > 0) {
                        List<Product> finalSpecificItems = specificItems;
                        warehouse.send(Messages.FullfilledOrderMessage.class, msg -> {
                            msg.orderContents = finalSpecificItems;
                            msg.warehouseName = warehouse.name;
                            msg.arrivalDate =
                                    order.warehouseOrder.shippingPreference ==
                                            ShippingPreference.EXPEDITED ?
                                            warehouse.getPrng().getNextInt(3) :
                                            warehouse.getPrng().getNextInt(7);
                        }).to(order.warehouseOrder.customerID);
                        warehouse.inventory.get(productID).removeAll(specificItems);
                    }
                });

                warehouse.writeData(demand);

                warehouse.totalInventory = warehouse.inventory.values().stream().mapToInt(List::size).sum();


                //send the demand for the step back to the production facility
                warehouse.getLinks(Links.WarehouseToFactoryLink.class).send(Messages.DemandMessage.class, (msg, link) -> {
                    msg.inventory = warehouse.inventory;
                    msg.demand = demand;
                });
            });


    @Override
    protected void receiveInboundTransport() {
        super.receiveInboundTransport();
    }

    @Override
    protected void processInboundLoadingBay() {
        pullArrivingTransport();
        unloadTransportCargo();
    }

    @Override
    protected void processOutboundLoadingBay() {
    }


    @Override
    protected void departTransport() {
    }


    public List<Integer> findEmptyBays(Map<Integer, TransportMessages.TransportMessage> loadingDocks) {
        List<Integer> emptyBays = new ArrayList<>();
        for (Map.Entry<Integer, TransportMessages.TransportMessage> bay : loadingDocks.entrySet()) {
            if (bay.getValue() == null) {
                emptyBays.add(bay.getKey());
            }
        }
        return emptyBays;
    }

    public void pullArrivingTransport() {
        List<TransportMessages.TransportMessage> arrivedTransport = inboundLoadingBay.transportQueue.entrySet()
                .stream()
                .filter(entry -> entry.getKey() == getContext().getTick())
                .map(Map.Entry::getValue)
                .collect(Collectors.toList());

        if (!arrivedTransport.isEmpty()) {
            List<Integer> emptyBays = findEmptyBays(inboundLoadingBay.loadingDocks);
            while (arrivedTransport.size() > 0 && emptyBays.size() > 0) {

                TransportMessages.TransportMessage t = arrivedTransport.remove(0);
                int emptyBay = emptyBays.remove(0);
                inboundLoadingBay.loadingDocks.put(emptyBay, t);
            }
        }
    }

    public void unloadTransportCargo() {
        List<TransportMessages.TransportMessage> processedtrucks = new ArrayList<>();
        inboundLoadingBay.loadingDocks.values().stream().filter(dock -> dock != null).forEach(t -> {
            List<Product> cargo = t.objectContents;
            cargo.stream().forEach(p -> {
                nbUnitsProcessed++;
                totalShipmentCost += p.weight * shipmentCostPeLb;
                inventory.computeIfAbsent(p.productID, k -> new ArrayList<>()).add(p);
            });
            processedtrucks.add(t);
            costPerUnit = totalShipmentCost / nbUnitsProcessed;
        });
        inboundLoadingBay.loadingDocks.entrySet().stream().forEach(bay -> {
            if (processedtrucks.contains(bay.getValue())) {
                bay.setValue(null);
            }
        });
    }

    public void writeData(Map<String, Integer> demand) {
        if (getContext().getTick() >= getGlobals().prodPredictionTime) {
            demand.entrySet().stream().forEach(entry -> {
                int inventory_p = inventory.get(entry.getKey()).size();
                int demand_p = entry.getValue();
                String id_p = entry.getKey();
                String newData = name + "," + id_p + "," + demand_p + "," + inventory_p + "," + getContext().getTick();
                writeToFile(newData);
            });
            List<String> noDemandProducts = getGlobals().availableProducts.keySet().stream()
                    .filter(key -> !demand.keySet().contains(key))
                    .collect(Collectors.toList());
            if (!noDemandProducts.isEmpty()) {
                noDemandProducts.stream().forEach(prod -> {
                    String newData = name + "," + prod + "," + 0 + "," + inventory.get(prod).size() + "," + getContext().getTick();
                    writeToFile(newData);
                });
            }
        }
    }

    public static void writeToFile(String content) {
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(FILE_NAME, true))) {
            writer.write(content);
            writer.newLine();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
